(* Copyright (C) 1994, Digital Equipment Corporation           *)
(* All rights reserved.                                        *)
(* See the file COPYRIGHT for a full description.              *)
(*                                                             *)
(* Last modified on Mon May 23 09:08:26 PDT 1994 by kalsow     *)

(* The code below makes the following NASTY assumption:
      ThreadF.ProcessStacks calls its argument twice for
      each thread -- the first time for the stack, the
      second time for its registers. *)

UNSAFE MODULE RTHeapStats;

IMPORT RT0, RT0u, RTCollector, RTModule, RTIO, RTHeapMap, RTHeapRep, RTMisc;
IMPORT RTType, RTStack, RTutils, RTProcedure, RTProcedureSRC;
IMPORT ThreadF, Cstring, Word, Text;

TYPE
  Info = RECORD
    module    : RT0.ModulePtr;
    stack_min : ADDRESS;
    stack_max : ADDRESS;
    location  : ADDRESS;
    ref       : ADDRESS;
    n_objects : INTEGER;
    n_bytes   : INTEGER;
  END;

TYPE
  InfoSet = RECORD
    count : INTEGER;
    info  : ARRAY [0..19] OF Info;
  END;

TYPE
  VisitStack = ARRAY [0..10000] OF ADDRESS;

CONST
  MapGrain = 2 * BYTESIZE (RT0.RefHeader);  (* = 1 bit in the map *)
  MapBitsPerHeapPage = RTHeapRep.BytesPerPage DIV MapGrain;
  MapWordsPerHeapPage = MapBitsPerHeapPage DIV BITSIZE (Word.T);

VAR
  units       : InfoSet;
  unit_roots  : InfoSet;
  stacks      : InfoSet;
  stack_roots : InfoSet;
  stack_pages : InfoSet;
  map         : UNTRACED REF ARRAY OF Word.T;
  heap_min    : ADDRESS;
  heap_max    : ADDRESS;
  visit       : Info;
  is_registers: BOOLEAN;
  visit_stack : UNTRACED REF VisitStack;
  top_of_stack: INTEGER;
  n_overflows : INTEGER;
  last_alloc  : ADDRESS;

PROCEDURE ReportReachable () =
  CONST MByte = 1024 * 1024;
  BEGIN
    (* allocate space for the stats *)
    visit_stack := NEW (UNTRACED REF VisitStack);
    map := NEW (UNTRACED REF ARRAY OF Word.T,
                 (RTHeapRep.p1 - RTHeapRep.p0) * MapWordsPerHeapPage);

    (* initialize the globals *)
    units.count       := 0;
    unit_roots.count  := 0;
    stacks.count      := 0;
    stack_roots.count := 0;
    stack_pages.count := 0;
    top_of_stack      := 0;
    n_overflows       := 0;

    (* freeze the world *)
    RTCollector.Disable ();
    RTHeapRep.Lock (); (* freeze the heap *)

    (* capture the heap limits *)
    heap_min  := LOOPHOLE (RTHeapRep.p0 * RTHeapRep.BytesPerPage, ADDRESS);
    heap_max  := LOOPHOLE (RTHeapRep.p1 * RTHeapRep.BytesPerPage, ADDRESS);

    Txt ("\nHEAP: ");
    Adr (heap_min);
    Txt (" .. ");
    Adr (heap_max);
    Txt (" => ");
    Int ((heap_max - heap_min) DIV MByte);
    Txt (".");
    Int ((heap_max - heap_min) * 10 DIV MByte MOD 10);
    Txt (" Mbytes\n");

    (* find the edge of the new space *)
    last_alloc := LOOPHOLE (NEW (REF INTEGER), ADDRESS);

    FOR i := 0 TO RTModule.Count() - 1 DO GetUnitStats (i); END;
    FOR i := 0 TO RTModule.Count() - 1 DO GetUnitRootStats (i); END;
    is_registers := FALSE;
    ThreadF.ProcessStacks (GetThreadStats);
    is_registers := FALSE;
    ThreadF.ProcessStacks (GetThreadRootStats);
    is_registers := FALSE;
    ThreadF.ProcessStacks (GetThreadPageStats);

    IF (n_overflows > 0) THEN
      Txt ("  ** warning: ");
      Int (n_overflows);
      Txt (" paths, longer than ");
      Int (NUMBER (VisitStack));
      Txt (" REFs, were truncated.\n");
    END;

    ReportUnits ();
    ReportUnitRoots ();

    ReportStacks ();
    ReportStackRoots ();
    ReportStackPages ();

    ReportStackPCs ();
    ReportStackRootPCs ();
    ReportStackPagePCs ();

    (* thaw the world *)
    DISPOSE (visit_stack);
    DISPOSE (map);
    RTHeapRep.Unlock (); (* unfreeze the heap *)
    RTCollector.Enable ();
  END ReportReachable;

(*------------------------------------------------------------ REF visits ---*)

PROCEDURE ResetVisitCounts () =
  BEGIN
    visit.n_objects := 0;
    visit.n_bytes   := 0;
    top_of_stack    := 0;
    RTMisc.Zero (ADR (map[0]), BYTESIZE (map^));
  END ResetVisitCounts;

PROCEDURE AddVisit (VAR s: InfoSet) =
  VAR n: INTEGER;
  BEGIN
    (* if the set isn't full, make room for this visit *)
    IF (s.count < NUMBER (s.info)) THEN
      s.info[s.count].n_bytes := -1;
      INC (s.count);
    END;

    (* find where to insert this visit *)
    n := s.count-1;
    WHILE (n >= 0) AND (s.info[n].n_bytes < visit.n_bytes) DO
      IF (n < LAST(s.info)) THEN s.info[n+1] := s.info[n]; END;
      DEC (n);
    END;
    INC (n);

    (* insert the new root *)
    IF (n < s.count) THEN  s.info[n] := visit;  END;
  END AddVisit;

PROCEDURE Visit (loc: ADDRESS) =
  BEGIN
    InnerVisit (loc);
    WHILE (top_of_stack > 0) DO
      DEC (top_of_stack);
      RTHeapMap.WalkRef (visit_stack[top_of_stack], InnerVisit);
    END;
  END Visit;

PROCEDURE InnerVisit (loc: ADDRESS) =
  CONST Mask = ADRSIZE (RT0.RefHeader) - 1; (* assume it's 2^k-1 for some k *)
  VAR ptr : UNTRACED REF ADDRESS := loc;
  VAR ref : ADDRESS := ptr^;
  VAR header: RTHeapMap.ObjectPtr;
  VAR cell, word, bit, mask, typecode: INTEGER;
  BEGIN
    header := ref - ADRSIZE (RT0.RefHeader);
    IF (heap_min <= ref) AND (ref < heap_max)
      AND (Word.And (LOOPHOLE(ref, INTEGER), Mask) = 0) THEN
      typecode := header.typecode;
      IF (0 < typecode) AND (typecode < RT0u.nTypes) THEN
        cell := (ref - heap_min) DIV MapGrain;
        word := cell DIV BITSIZE (Word.T);
        bit  := cell - word * BITSIZE (Word.T);
        mask := Word.LeftShift (1, bit);
        IF (Word.And (mask, map[word]) = 0) THEN
          (* this is a new ref... *)
          map[word] := Word.Or (mask, map[word]);
          INC (visit.n_objects);
          INC (visit.n_bytes, DataSize (header) + BYTESIZE (RT0.RefHeader));
          IF (top_of_stack < NUMBER (VisitStack)) THEN
            visit_stack [top_of_stack] := header;
            INC (top_of_stack);
          ELSE
            INC (n_overflows);
          END;
        END;
      END;
    END;
  END InnerVisit;

PROCEDURE DataSize (h: RTHeapMap.ObjectPtr): CARDINAL =
  VAR
    res : INTEGER;
    tc  : RT0.Typecode := h.typecode;
    def : RT0.TypeDefn;
  BEGIN
    IF tc = RTHeapRep.Fill_1_type THEN RETURN 0; END;
    IF tc = RTHeapRep.Fill_N_type THEN
      res := LOOPHOLE(h + ADRSIZE(RT0.RefHeader), UNTRACED REF INTEGER)^;
      RETURN res - BYTESIZE(RT0.RefHeader);
    END;
    def := RTType.Get (tc);
    IF def.nDimensions = 0 THEN
      (* the typecell datasize tells the truth *)
      RETURN def.dataSize;
    END;
    (* ELSE, the referent is an open array; it has the following layout:
|         pointer to the elements (ADDRESS)
|         size 1
|         ....
|         size n
|         optional padding
|         elements
|         ....
       where n is the number of open dimensions (given by the definition)
       and each size is the number of elements along the dimension *)
    VAR
      sizes: UNTRACED REF INTEGER := h + ADRSIZE(RT0.RefHeader)
                                       + ADRSIZE(ADDRESS);  (* ^ elt pointer*)
    BEGIN
      res := 1;
      FOR i := 0 TO def.nDimensions - 1 DO
        res := res * sizes^;
        INC(sizes, ADRSIZE(sizes^));
      END;
      res := res * def.elementSize;
    END;
    res := RTMisc.Upper(res + def.dataSize, BYTESIZE(RT0.RefHeader));
    RETURN res;
  END DataSize;

PROCEDURE TypeName (ref: ADDRESS): TEXT =
  VAR header: RTHeapMap.ObjectPtr;
  VAR typecode: INTEGER;
  BEGIN
    header := ref - ADRSIZE (RT0.RefHeader);
    IF (heap_min <= ref) AND (ref < heap_max) THEN
      typecode := header.typecode;
      IF (0 < typecode) AND (typecode < RT0u.nTypes) THEN
        RETURN RTutils.TypecodeName (typecode);
      END;
    END;
    RETURN "?";
  END TypeName;

(*----------------------------------------------------------------- units ---*)

PROCEDURE GetUnitStats (n: CARDINAL) =
  BEGIN
    visit.module     := RTModule.Get (n);
    visit.stack_min  := NIL;
    visit.stack_max  := NIL;
    visit.location   := NIL;
    visit.ref        := NIL;
    ResetVisitCounts ();
    RTHeapMap.WalkModuleGlobals (Visit, n);
    AddVisit (units);
  END GetUnitStats;

PROCEDURE GetUnitRootStats (n: CARDINAL) =
  BEGIN
    visit.module     := RTModule.Get (n);
    visit.stack_min  := NIL;
    visit.stack_max  := NIL;
    visit.location   := NIL;
    visit.ref        := NIL;
    RTHeapMap.WalkModuleGlobals (VisitRoot, n);
  END GetUnitRootStats;

PROCEDURE VisitRoot (root: ADDRESS) =
  VAR p: UNTRACED REF ADDRESS := root;
  BEGIN
    visit.location := root;
    visit.ref      := p^;
    ResetVisitCounts ();
    Visit (root);
    AddVisit (unit_roots);
  END VisitRoot;

(*--------------------------------------------------------------- threads ---*)

PROCEDURE GetThreadStats (start, stop: ADDRESS) =
  VAR fp := start;  p: ADDRESS;  page: INTEGER;
  BEGIN
    IF (NOT is_registers) THEN
      visit.module     := NIL;
      visit.stack_min  := start;
      visit.stack_max  := stop;
      visit.location   := NIL;
      visit.ref        := NIL;
      ResetVisitCounts ();
    END;

    (* scan the stack or registers *)
    WHILE fp <= stop DO
      p := LOOPHOLE(fp, UNTRACED REF ADDRESS)^;
      IF heap_min <= p AND p < heap_max THEN
        page := (p - heap_min) DIV RTHeapRep.BytesPerPage;
        IF RTHeapRep.desc[page].space = RTHeapRep.Space.Current THEN
          VisitPage (page);
        END;
      END;
      INC(fp, RTStack.PointerAlignment);
    END;

    IF (is_registers) THEN AddVisit (stacks); END;
    is_registers := NOT is_registers;
  END GetThreadStats;

PROCEDURE GetThreadRootStats (start, stop: ADDRESS) =
  VAR fp := start;  p: ADDRESS;  page: INTEGER;
  BEGIN
    IF (is_registers) THEN
      visit.location   := NIL;
    ELSE
      visit.module     := NIL;
      visit.stack_min  := start;
      visit.stack_max  := stop;
    END;

    (* scan the stack *)
    WHILE fp <= stop DO
      p := LOOPHOLE(fp, UNTRACED REF ADDRESS)^;
      IF heap_min <= p AND p < heap_max THEN
        page := (p - heap_min) DIV RTHeapRep.BytesPerPage;
        IF RTHeapRep.desc[page].space = RTHeapRep.Space.Current THEN
          IF (NOT is_registers) THEN visit.location := fp; END;
          visit.ref      := p;
          ResetVisitCounts ();
          Visit (fp);
          AddVisit (stack_roots);
        END;
      END;
      INC(fp, RTStack.PointerAlignment);
    END;

    is_registers := NOT is_registers;
  END GetThreadRootStats;

PROCEDURE GetThreadPageStats (start, stop: ADDRESS) =
  VAR fp := start;  p: ADDRESS;  page: INTEGER;
  BEGIN
    IF (is_registers) THEN
      visit.location   := NIL;
    ELSE
      visit.module     := NIL;
      visit.stack_min  := start;
      visit.stack_max  := stop;
    END;

    (* scan the stack *)
    WHILE fp <= stop DO
      p := LOOPHOLE(fp, UNTRACED REF ADDRESS)^;
      IF heap_min <= p AND p < heap_max THEN
        page := (p - heap_min) DIV RTHeapRep.BytesPerPage;
        IF RTHeapRep.desc[page].space = RTHeapRep.Space.Current THEN
          IF (NOT is_registers) THEN visit.location := fp; END;
          visit.ref      := p;
          ResetVisitCounts ();
          VisitPage (page);
          AddVisit (stack_pages);
        END;
      END;
      INC(fp, RTStack.PointerAlignment);
    END;

    is_registers := NOT is_registers;
  END GetThreadPageStats;

PROCEDURE VisitPage (page: INTEGER) =
  VAR start, stop: ADDRESS;  h: RTHeapMap.ObjectPtr;  ref: ADDRESS;
  BEGIN
    (* find the address limits of this "page" *)
    WHILE (page > 0) AND (RTHeapRep.desc[page].continued) DO DEC (page); END;
    start := heap_min + page * RTHeapRep.BytesPerPage;
    REPEAT
      INC (page);
    UNTIL (page >= RTHeapRep.p1-RTHeapRep.p0)
       OR (NOT RTHeapRep.desc[page].continued);
    stop := heap_min + page * RTHeapRep.BytesPerPage;

    IF (start <= last_alloc) AND (last_alloc < stop) THEN
      (* we're on the allocator's partial page... *)
      stop := last_alloc;
    END;

    (* visit each object on the page *)
    h := start;
    WHILE (h < stop) AND (h.typecode # 0) DO
      ref := h + ADRSIZE (RT0.RefHeader);
      Visit (ADR (ref));
      INC (h, DataSize (h) + ADRSIZE (RT0.RefHeader));
    END;
  END VisitPage;

(*--------------------------------------------------------------- reports ---*)

PROCEDURE ReportUnits () =
  BEGIN
    Txt ("\nModule globals:\n");
    Txt (" # objects   # bytes  unit\n");
    Txt (" ---------  --------  -----------------\n");
    FOR i := 0 TO units.count-1 DO
      WITH m = units.info[i] DO
        IF (m.n_bytes > 0) THEN
          Int (m.n_objects, 10);
          Int (m.n_bytes, 10);
          Txt ("  ");
          Str (PathTail (m.module.file));
          Txt ("\n");
        END;
      END;
    END;
  END ReportUnits;

PROCEDURE ReportUnitRoots () =
  BEGIN
    Txt ("\nGlobal variable roots:\n");
    Txt (" # objects   # bytes         ref type                location\n");
    Txt (" ---------  --------  ---------- -----------------   ------------------------\n");
    FOR i := 0 TO unit_roots.count-1 DO
      WITH r = unit_roots.info[i] DO
        IF (r.n_bytes > 0) THEN
          Int (r.n_objects, 10);
          Int (r.n_bytes, 10);
          Txt ("  ");
          Adr (r.ref);
          Txt (" ");
          Txt (TypeName (r.ref), 18);
          Txt ("  ");
          Str (PathTail (r.module.file));
          Txt (" + ");
          Int (r.location - r.module);
          Txt ("\n");
        END;
      END;
    END;
  END ReportUnitRoots;

PROCEDURE ReportStacks () =
  BEGIN
    Txt ("\nThread stacks (conservative page scan):\n");
    Txt (" # objects   # bytes  stack bounds\n");
    Txt (" ---------  --------  ------------------------\n");
    FOR i := 0 TO stacks.count-1 DO
      WITH t = stacks.info[i] DO
        IF (t.n_bytes > 0) THEN
          Int (t.n_objects, 10);
          Int (t.n_bytes, 10);
          Txt ("  [");
          Adr (t.stack_min);
          Txt ("..");
          Adr (t.stack_max);
          Txt ("]\n");
        END;
      END;
    END;
  END ReportStacks;

PROCEDURE ReportStackPCs () =
  VAR stack_mid: ADDRESS;
  BEGIN
    Txt ("\nThread stack PCs:\n");
    Txt ("     SP         PC     procedure\n");
    Txt ("----------  ---------  ------------------------\n");
    FOR i := 0 TO stacks.count-1 DO
      WITH t = stacks.info[i] DO
        IF (t.n_bytes > 0) THEN
          Txt ("                          [");
          Adr (t.stack_min);
          Txt ("..");
          Adr (t.stack_max);
          Txt ("]\n");
          stack_mid := t.stack_min + (t.stack_max - t.stack_min) DIV 2;
          ReportPCs (t.stack_min, stack_mid, 5, +1);
          Txt ("  ...\n");
          ReportPCs (t.stack_max, stack_mid, 5, -1);
          Txt ("\n");
        END;
      END;
    END;
  END ReportStackPCs;

PROCEDURE ReportStackRoots () =
  BEGIN
    Txt ("\nThread stack roots (optimistic):\n");
    ReportStackInfo (stack_roots);
  END ReportStackRoots;

PROCEDURE ReportStackPages () =
  BEGIN
    Txt ("\nThread stack roots (conservative page scan):\n");
    ReportStackInfo (stack_pages);
  END ReportStackPages;

PROCEDURE ReportStackInfo (READONLY s: InfoSet) =
  BEGIN
    Txt (" # objects   # bytes         ref type                location\n");
    Txt (" ---------  --------  ---------- -----------------   ------------------------\n");
    FOR i := 0 TO s.count-1 DO
      WITH r = s.info[i] DO
        IF (r.n_bytes > 0) THEN
          Int (r.n_objects, 10);
          Int (r.n_bytes, 10);
          Txt ("  ");
          Adr (r.ref);
          Txt (" ");
          Txt (TypeName (r.ref), 18);
          Txt ("  ");
          IF (r.location # NIL) THEN
            Txt ("sp+");
            Int (r.location - r.stack_min);
          ELSE
            Txt ("register");
          END;
          Txt (" in [");
          Adr (r.stack_min);
          Txt ("..");
          Adr (r.stack_max);
          Txt ("]");
          Txt ("\n");
        END;
      END;
    END;
  END ReportStackInfo;

PROCEDURE ReportStackRootPCs () =
  BEGIN
    Txt ("\nThread stack root PCs (optimistic):\n");
    ReportStackInfoPCs (stack_roots);
  END ReportStackRootPCs;

PROCEDURE ReportStackPagePCs () =
  BEGIN
    Txt ("\nThread stack root PCs (conservative page scan):\n");
    ReportStackInfoPCs (stack_pages);
  END ReportStackPagePCs;

PROCEDURE ReportStackInfoPCs (READONLY s: InfoSet) =
  VAR m: ARRAY [0..LAST(s.info)] OF INTEGER;  xx, xy: ADDRESS;
  BEGIN
    (* first, sort the set by location *)
    m[0] := 0;
    FOR i := 1 TO s.count-1 DO
      VAR j := i-1; key := s.info[i].location;  BEGIN
        WHILE (j >= 0) AND (s.info[m[j]].location > key) DO
          m[j+1] := m[j];
          DEC (j);
        END;
        m[j+1] := i;
      END;
    END;

    Txt ("     SP         PC     location\n");
    Txt ("----------  ---------  ------------------------\n");
    FOR i := 0 TO s.count-1 DO
      WITH r = s.info[m[i]] DO
        IF (r.n_bytes > 0) AND (r.location # NIL) THEN
          xy := r.location - RTStack.PointerAlignment;
          IF (i = 0) OR (s.info[m[i-1]].location = NIL) OR
            (s.info[m[i-1]].stack_min # r.stack_min) THEN
            (* this is the first entry on this stack *)
            xx := xy + (r.stack_min - xy) DIV 2;
            Adr (r.stack_min, 10);
            Txt ("              [");
            Adr (r.stack_min);
            Txt ("..");
            Adr (r.stack_max);
            Txt ("]\n");
            ReportPCs (r.stack_min, xx, 3, +1);
          ELSE
            xx := xy + (s.info[m[i-1]].location - xy) DIV 2;
          END;
          Txt ("...\n");
          ReportPCs (xy, xx, 3, -1);
          Adr (r.location, 10);
          Txt ("                ");
          Txt ("sp+");
          Int (r.location - r.stack_min);
          Txt ("\n");
          xy := r.location + RTStack.PointerAlignment;
          IF (i = s.count-1) OR (s.info[m[i+1]].location = NIL) OR
            (s.info[m[i+1]].stack_min # r.stack_min) THEN
            (* this is the last entry on this stack *)
            xx := xy + (r.stack_max - xy) DIV 2;
            ReportPCs (xy, xx, 3, +1);
            Txt ("...\n");
            ReportPCs (r.stack_max, xx, 3, -1);
            Adr (r.stack_max, 10);
            Txt ("\n\n");
          ELSE
            xx := xy + (s.info[m[i+1]].location - xy) DIV 2;
            ReportPCs (xy, xx, 3, +1);
          END;
        END;
      END;
    END;
  END ReportStackInfoPCs;

(*--------------------------------------------------------------- PC info ---*)

CONST
  Max_proc = 4096;  (* good enough for 99% of the procedures *)

TYPE
  PCInfo = RECORD
    loc  : ADDRESS;
    pc   : ADDRESS;
    proc : RTProcedure.Proc;
    file : RTProcedureSRC.Name;
    name : RTProcedureSRC.Name;
  END;

PROCEDURE ReportPCs (start, stop: ADDRESS;  max, dir: INTEGER) =
  VAR
    x   : ARRAY [0..9] OF PCInfo;
    cnt := FindPCs (start, stop, SUBARRAY (x, 0, MIN (max, NUMBER(x))));
    a, b: INTEGER;
  BEGIN
    a := 0;
    b := cnt-1;
    IF (dir < 0) THEN a := b; b := 0; END;
    FOR i := a TO b BY dir DO
      WITH p = x[i] DO
        Adr (p.loc, 10);
        Txt (" ");
        Adr (p.pc, 10);
        Txt ("  ");
        Str (p.name);
        IF (p.pc # p.proc) THEN
          Txt (" + ");
          Int (p.pc - p.proc);
        END;
        IF (p.file # NIL) THEN
          Txt (" in ");
          Str (PathTail (p.file));
        END;
        Txt ("\n");
      END;
    END;
  END ReportPCs;

PROCEDURE FindPCs (start, stop: ADDRESS;  VAR x: ARRAY OF PCInfo): INTEGER =
  VAR n := 0;  fp: UNTRACED REF ADDRESS := start;
  BEGIN
    IF (start < stop) THEN
      WHILE (fp < stop) AND (n < NUMBER (x)) DO
        WITH p = x[n] DO
          p.loc := fp;
          p.pc  := fp^;
          RTProcedureSRC.FromPC (p.pc, p.proc, p.file, p.name);
          IF (p.proc # NIL) AND (p.pc - p.proc < Max_proc) THEN INC (n) END;
        END;
        INC (fp, RTStack.PointerAlignment);
      END;
    ELSE
      WHILE (fp > stop) AND (n < NUMBER (x)) DO
        WITH p = x[n] DO
          p.loc := fp;
          p.pc := fp^;
          RTProcedureSRC.FromPC (p.pc, p.proc, p.file, p.name);
          IF (p.proc # NIL) AND (p.pc - p.proc < Max_proc) THEN INC (n) END;
        END;
        DEC (fp, RTStack.PointerAlignment);
      END;
    END;
    RETURN n;
  END FindPCs;
     
(*--------------------------------------------------------- low-level I/O ---*)

PROCEDURE PathTail (a: ADDRESS): ADDRESS =
  VAR p0 : UNTRACED REF CHAR := a;  p := p0;
  BEGIN
    IF (p0 = NIL) THEN RETURN NIL END;
    WHILE (p^ # '\000') DO
      IF (p^ = '/') THEN p0 := p + ADRSIZE (p^); END;
      INC (p, ADRSIZE (p^));
    END;
    RETURN p0;
  END PathTail;

PROCEDURE Str (s: ADDRESS) =
  BEGIN
    IF (s = NIL) THEN RETURN END;
    RTIO.PutChars (RTIO.stderr, s, Cstring.strlen (s));
  END Str;

PROCEDURE Txt (t: TEXT;  width := 0) =
  VAR len := Text.Length (t);
  BEGIN
    RTIO.PutText (RTIO.stderr, t);
    WHILE (len < width) DO
      RTIO.PutChar (RTIO.stderr, ' ');
      INC (len);
    END;
  END Txt;

PROCEDURE Int (i: INTEGER;  width: CARDINAL := 0) =
  BEGIN
    RTIO.PutInt (RTIO.stderr, i, width)
  END Int;

PROCEDURE Hex (i: INTEGER;  width: CARDINAL := 0) =
  BEGIN
    RTIO.PutHexa (RTIO.stderr, i, width)
  END Hex;

PROCEDURE Adr (a: ADDRESS; width: CARDINAL := 0) =
  BEGIN
    Hex (LOOPHOLE (a, INTEGER), width)
  END Adr;

BEGIN
END RTHeapStats.

   
   
